@@ -17,18 +17,20 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#include <stdlib.h>
 #include <stdbool.h>
 
+#include "gl_list.h"
+
 struct kvelem {
     char *key;
     void *value;
 };
 
-size_t kvhash_hasher (const void *data, size_t n);
-bool kvhash_comparator (const void *a, const void *b);
-Hash_table *kvhash_initialize (size_t candidate, const Hash_tuning *tuning,
-			       Hash_data_freer data_freer);
-bool kvhash_exists (const Hash_table *table, const char *key);
-void *kvhash_lookup (const Hash_table *table, const char *key);
-void *kvhash_insert (Hash_table *table, const char *key, const void *value);
-void *kvhash_delete (Hash_table *table, const char *key);
+void kvhash_dispose (void *data);
+void kvhash_dispose_plain (void *data);
+gl_list_t kvhash_initialize (gl_listelement_dispose_fn dispose_fn);
+bool kvhash_exists (gl_list_t list, const char *key);
+void *kvhash_lookup (gl_list_t list, const char *key);
+void *kvhash_insert (gl_list_t list, const char *key, const void *value);
+void *kvhash_delete (gl_list_t list, const char *key);
