@@ -46,8 +46,7 @@
 #include <pipeline.h>
 
 #include "argp.h"
-#include "gl_xlist.h"
-#include "hash.h"
+#include "gl_list.h"
 #include "xalloc.h"
 #include "xvasprintf.h"
 
@@ -57,9 +56,6 @@
 #include "kvhash.h"
 #include "paths.h"
 
-#define HASH_FOR_EACH(iter, hash) \
-    for (iter = hash_get_first (hash); iter; iter = hash_get_next (hash, iter))
-
 char *program_name;
 
 static int test = 0;
@@ -67,7 +63,7 @@ static int test = 0;
 static char *path_register, *path_status;
 static char *run_detectors;
 
-Hash_table *formats;
+gl_list_t formats;
 
 static inline bool exists (const char *name)
 {
@@ -114,14 +110,14 @@ static int rename_mv (const char *source, const char *dest)
     if (!rename (source, dest))
 	return 1;
 
-    mv = pipeline_new_command_args ("mv", source, dest, NULL);
+    mv = pipeline_new_command_args ("mv", source, dest, (void *) 0);
     return pipeline_run (mv) == 0;
 }
 
-static Hash_table *get_import (const char *name)
+static gl_list_t get_import (const char *name)
 {
     FILE *import_file;
-    Hash_table *import;
+    gl_list_t import;
     char *line = NULL;
     ssize_t len;
     size_t n;
@@ -131,7 +127,8 @@ static Hash_table *get_import (const char *name)
 	warning_err ("unable to open %s", name);
 	return NULL;
     }
-    import = kvhash_initialize (10, NULL, free);
+    import = kvhash_initialize
+	((gl_listelement_dispose_fn) kvhash_dispose_plain);
 
     while ((len = getline (&line, &n, import_file)) != -1) {
 	char *space, *key, *value, *p, *q;
@@ -163,92 +160,40 @@ static Hash_table *get_import (const char *name)
     return import;
 }
 
+/* Check whether binfmt_misc is enabled.
+ *
+ * This lets us avoids mounting /proc/sys/fs/binfmt_misc when update-binfmts
+ * --import is run in a chroot, which in turn avoids possible busy-mount
+ * problems when cleaning up the chroot.
+ *
+ * This isn't an entirely reliable test: in particular, if systemd is built
+ * with ENABLE_BINFMT, it sets up /proc/sys/fs/binfmt_misc to be
+ * automounted, so just checking for the existence of
+ * /proc/sys/fs/binfmt_misc/register will cause it to be mounted.  However,
+ * if systemd is running then you're almost certainly in an environment that
+ * knows how to clean up the mount again, so we can probably afford to
+ * ignore this.
+ */
+static int is_enabled (void)
+{
+    return is_file (path_register);
+}
+
 /* Loading and unloading logic, which should cope with the various ways this
  * has been implemented.
  */
 
-enum binfmt_style {
-    BINFMT_PROCFS,
-    BINFMT_FILESYSTEM
-};
-
-static enum binfmt_style get_binfmt_style (void)
-{
-    FILE *fs;
-    char *line = NULL;
-    ssize_t len;
-    size_t n;
-
-    fs = fopen ("/proc/filesystems", "r");
-    if (!fs) {
-	/* Weird.  Assume procfs. */
-	warning_err ("unable to open /proc/filesystems");
-	return BINFMT_PROCFS;
-    }
-    while ((len = getline (&line, &n, fs)) != -1) {
-	if (len && line[len - 1] == '\n')
-	    line[--len] = 0;
-
-	if (len >= sizeof "binfmt_misc" &&
-	    /* sizeof - 1 (NUL) + 1 (check preceding character) */
-	    isspace ((int) line[len - sizeof "binfmt_misc"]) &&
-	    !strcmp (line + len - sizeof "binfmt_misc" + 1, "binfmt_misc")) {
-	    /* Linux gained a binfmt_misc filesystem type at some point,
-	     * which needs to be mounted separately.
-	     */
-	    free (line);
-	    fclose (fs);
-	    return BINFMT_FILESYSTEM;
-	}
-	free (line);
-	line = NULL;
-    }
-    fclose (fs);
-    /* The traditional (obsolete) interface. */
-    return BINFMT_PROCFS;
-}
-
 static int load_binfmt_misc (void)
 {
-    enum binfmt_style style;
-
     if (test) {
 	printf ("load binfmt_misc\n");
 	return 1;
     }
 
-    style = get_binfmt_style ();
-    if (style == BINFMT_PROCFS && !is_file (path_register)) {
-	int failed = 0;
-
-	if (access ("/sbin/modprobe", X_OK))
-	    failed = 1;
-	else {
-	    pipeline *modprobe = pipeline_new_command_args
-		("/sbin/modprobe", "-q", "binfmt_misc", NULL);
-
-	    if (pipeline_run (modprobe))
-		failed = 1;
-	}
-
-	if (failed) {
-	    warning ("Couldn't load the binfmt_misc module.");
-	    return 0;
-	}
-    }
-
-    if (!is_directory (procdir)) {
-	warning ("binfmt_misc module seemed to be loaded, but no %s "
-		 "directory!  Giving up.", procdir);
-	return 0;
-    }
-
-    /* Find out what the style looks like now. */
-    style = get_binfmt_style ();
-    if (style == BINFMT_FILESYSTEM && !is_file (path_register)) {
+    if (!is_enabled ()) {
 	pipeline *mount = pipeline_new_command_args
 	    ("/bin/mount", "-t", "binfmt_misc", "-o", "nodev,noexec,nosuid",
-	     "binfmt_misc", procdir, NULL);
+	     "binfmt_misc", procdir, (void *) 0);
 
 	if (pipeline_run (mount)) {
 	    warning ("Couldn't mount the binfmt_misc filesystem on %s.",
@@ -257,7 +202,7 @@ static int load_binfmt_misc (void)
 	}
     }
 
-    if (is_file (path_register)) {
+    if (is_enabled ()) {
 	FILE *status_file = fopen (path_status, "w");
 	if (status_file) {
 	    fprintf (status_file, "1\n");
@@ -274,44 +219,33 @@ static int load_binfmt_misc (void)
 
 static int unload_binfmt_misc (void)
 {
-    enum binfmt_style style = get_binfmt_style ();
+#ifdef __linux__
+    struct statfs stfs;
+#endif /* __linux__ */
+    pipeline *umount;
 
     if (test) {
-	printf ("unload binfmt_misc (%s)\n",
-		style == BINFMT_PROCFS ? "procfs" : "filesystem");
+	printf ("unload binfmt_misc\n");
 	return 1;
     }
 
-    if (style == BINFMT_FILESYSTEM) {
 #ifdef __linux__
-	struct statfs stfs;
-#endif /* __linux__ */
-	pipeline *umount;
-
-#ifdef __linux__
-	if (statfs (procdir, &stfs) == 0 && stfs.f_type == FUSE_SUPER_MAGIC) {
-	    /* Probably running within test suite.  Try fusermount -u. */
-	    umount = pipeline_new_command_args
-		("fusermount", "-u", procdir, NULL);
-	    if (!pipeline_run (umount))
-		return 1;
-	}
+    if (statfs (procdir, &stfs) == 0 && stfs.f_type == FUSE_SUPER_MAGIC) {
+        /* Probably running within test suite.  Try fusermount -u. */
+        umount = pipeline_new_command_args
+            ("fusermount", "-u", procdir, NULL);
+        if (!pipeline_run (umount))
+            return 1;
+    }
 #endif /* __linux__ */
 
-	umount = pipeline_new_command_args ("/bin/umount", procdir, NULL);
-	if (pipeline_run (umount)) {
-	    warning ("Couldn't unmount the binfmt_misc filesystem from %s.",
-		     procdir);
-	    return 0;
-	}
+    umount = pipeline_new_command_args ("/bin/umount", procdir, (void *) 0);
+    if (pipeline_run (umount)) {
+        warning ("Couldn't unmount the binfmt_misc filesystem from %s.",
+                 procdir);
+        return 0;
     }
 
-    /* We used to try to unload the kernel module as well, but it seems that
-     * it doesn't always unload properly (http://bugs.debian.org/155570) and
-     * in any case it means that strictly speaking we have to remember if
-     * the module was loaded when we started.  Since it's not actually
-     * important, we now just don't bother.
-     */
     return 1;
 }
 
@@ -391,7 +325,8 @@ static int act_enable (const char *name)
 
 	need_detector = binfmt->detector && *binfmt->detector;
 	if (!need_detector) {
-	    struct kvelem *format_iter;
+	    gl_list_iterator_t format_iter;
+	    const void *format_elem;
 
 	    /* Scan the format database to see if anything else uses the
 	     * same spec as us. If so, assume that we need a detector,
@@ -401,14 +336,17 @@ static int act_enable (const char *name)
 	     */
 	    load_all_formats (1);
 
-	    HASH_FOR_EACH (format_iter, formats) {
-		if (!strcmp (format_iter->key, name))
+	    format_iter = gl_list_iterator (formats);
+	    while (gl_list_iterator_next (&format_iter, &format_elem, NULL)) {
+		const struct kvelem *format_kvelem = format_elem;
+		if (!strcmp (format_kvelem->key, name))
 		    continue;
-		if (binfmt_equals (binfmt, format_iter->value)) {
+		if (binfmt_equals (binfmt, format_kvelem->value)) {
 		    need_detector = 1;
 		    break;
 		}
 	    }
+	    gl_list_iterator_free (&format_iter);
 	}
 	/* Fake the interpreter if we need a userspace detector program. */
 	interpreter = need_detector ? run_detectors : binfmt->interpreter;
@@ -421,6 +359,12 @@ static int act_enable (const char *name)
 	fix_binary =
 		(binfmt->fix_binary && !strcmp (binfmt->fix_binary, "yes"))
 		? "F" : "";
+	if (need_detector && *fix_binary) {
+	    warning_err ("unable to enable binary format %s: another format "
+			 "with the same magic already exists and this is a "
+			 "fix-binary format", name);
+	    return 0;
+	}
 	regstring = xasprintf (":%s:%c:%s:%s:%s:%s:%s%s%s\n",
 			       name, type, binfmt->offset, binfmt->magic,
 			       binfmt->mask, interpreter,
@@ -446,15 +390,19 @@ static int act_enable (const char *name)
 	return 1;
     } else {
 	int worked = 1;
-	struct kvelem *format_iter;
+	gl_list_iterator_t format_iter;
+	const void *format_elem;
 
 	load_all_formats (0);
-	HASH_FOR_EACH (format_iter, formats) {
-	    procdir_name = xasprintf ("%s/%s", procdir, format_iter->key);
+	format_iter = gl_list_iterator (formats);
+	while (gl_list_iterator_next (&format_iter, &format_elem, NULL)) {
+	    const struct kvelem *format_kvelem = format_elem;
+	    procdir_name = xasprintf ("%s/%s", procdir, format_kvelem->key);
 	    if (!exists (procdir_name))
-		worked &= act_enable (format_iter->key);
+		worked &= act_enable (format_kvelem->key);
 	    free (procdir_name);
 	}
+	gl_list_iterator_free (&format_iter);
 	return worked;
     }
 }
@@ -514,21 +462,27 @@ static int act_disable (const char *name)
 	return 1;
     } else {
 	int worked = 1;
-	struct kvelem *format_iter;
+	gl_list_iterator_t format_iter;
+	const void *format_elem;
 
 	load_all_formats (0);
-	HASH_FOR_EACH (format_iter, formats) {
-	    char *procdir_id = xasprintf ("%s/%s", procdir, format_iter->key);
+	format_iter = gl_list_iterator (formats);
+	while (gl_list_iterator_next (&format_iter, &format_elem, NULL)) {
+	    const struct kvelem *format_kvelem = format_elem;
+	    char *procdir_id = xasprintf ("%s/%s", procdir,
+					  format_kvelem->key);
 	    if (exists (procdir_id))
-		worked &= act_disable (format_iter->key);
+		worked &= act_disable (format_kvelem->key);
 	    free (procdir_id);
 	}
+	gl_list_iterator_free (&format_iter);
 	unload_binfmt_misc (); /* ignore errors here */
 	return worked;
     }
 }
 
-static int act_install (const char *name, const struct binfmt *binfmt)
+static int act_install (const char *name, const struct binfmt *binfmt,
+			int always_update_kernel)
 {
     char *admindir_name;
 
@@ -559,10 +513,12 @@ static int act_install (const char *name, const struct binfmt *binfmt)
      */
     admindir_name = xasprintf ("%s/%s", admindir, name);
     if (is_file (admindir_name)) {
-	if (!act_disable (name)) {
-	    warning ("unable to disable binary format %s", name);
-	    free (admindir_name);
-	    return 0;
+	if (always_update_kernel || is_enabled ()) {
+	    if (!act_disable (name)) {
+		warning ("unable to disable binary format %s", name);
+		free (admindir_name);
+		return 0;
+	    }
 	}
     }
 
@@ -587,9 +543,11 @@ static int act_install (const char *name, const struct binfmt *binfmt)
     }
     free (admindir_name);
     kvhash_insert (formats, name, binfmt);
-    if (!act_enable (name)) {
-	warning ("unable to enable binary format %s", name);
-	return 0;
+    if (always_update_kernel || is_enabled ()) {
+	if (!act_enable (name)) {
+	    warning ("unable to enable binary format %s", name);
+	    return 0;
+	}
     }
     return 1;
 }
@@ -597,7 +555,8 @@ static int act_install (const char *name, const struct binfmt *binfmt)
 /* Remove a binary format from binfmt-support's database.  Attempt to
  * disable the format in the kernel first.
  */
-static int act_remove (const char *name, const char *package)
+static int act_remove (const char *name, const char *package,
+		       int always_update_kernel)
 {
     char *admindir_name;
 
@@ -627,10 +586,12 @@ static int act_remove (const char *name, const char *package)
 	    return 1;
 	}
     }
-    if (!act_disable (name)) {
-	warning ("unable to disable binary format %s", name);
-	free (admindir_name);
-	return 0;
+    if (always_update_kernel || is_enabled ()) {
+	if (!act_disable (name)) {
+	    warning ("unable to disable binary format %s", name);
+	    free (admindir_name);
+	    return 0;
+	}
     }
     if (test)
 	printf ("remove %s", admindir_name);
@@ -654,7 +615,7 @@ static int act_import (const char *name)
     if (name) {
 	const char *slash, *id;
 	char *path;
-	Hash_table *import;
+	gl_list_t import;
 	const char *interpreter;
 
 	slash = strrchr (name, '/');
@@ -674,7 +635,7 @@ static int act_import (const char *name)
 	}
 
 	import = get_import (path);
-	if (!import || !hash_get_n_entries (import)) {
+	if (!import || !gl_list_size (import)) {
 	    warning ("couldn't find information about '%s' to import", id);
 	    free (path);
 	    return 0;
@@ -709,7 +670,7 @@ static int act_import (const char *name)
 	    warning ("%s: no executable %s found, but continuing anyway as "
 		     "you request", path, interpreter);
 
-	act_install (id, binfmt_new (path, import));
+	act_install (id, binfmt_new (path, import), 0);
 	free (path);
 	return 1;
     } else {
@@ -737,6 +698,93 @@ static int act_import (const char *name)
     }
 }
 
+/* Unimport an old format file that was previously imported into
+ * binfmt-support's database.  This is intended for use by packaging
+ * systems.
+ */
+static int act_unimport (const char *name)
+{
+    if (name) {
+	const char *slash, *id;
+	char *path;
+	gl_list_t import;
+	const char *package;
+
+	slash = strrchr (name, '/');
+	if (slash) {
+	    id = slash + 1;
+	    path = xstrdup (name);
+	} else {
+	    id = name;
+	    path = xasprintf ("%s/%s", importdir, name);
+	}
+
+	if (!strcmp (id, ".") || !strcmp (id, "..") ||
+	    !strcmp (id, "register") || !strcmp (id, "status")) {
+	    warning ("binary format name '%s' is reserved", id);
+	    free (path);
+	    return 0;
+	}
+
+	import = get_import (path);
+	if (!import || !gl_list_size (import)) {
+	    warning ("couldn't find information about '%s' to unimport", id);
+	    free (path);
+	    return 0;
+	}
+
+	load_format (id, 1);
+	if (kvhash_exists (formats, id)) {
+	    const struct binfmt *binfmt = kvhash_lookup (formats, id);
+	    if (!strcmp (binfmt->package, ":")) {
+		/* Installed version was installed manually, so don't
+		 * unimport it.
+		 */
+		warning ("preserving local changes to %s", id);
+		free (path);
+		return 1;
+	    }
+	} else {
+	    /* Nothing to do. */
+	    free (path);
+	    return 1;
+	}
+
+	package = kvhash_lookup (import, "package");
+	if (!package) {
+	    warning ("%s: required 'package' line missing", path);
+	    free (path);
+	    return 0;
+	}
+
+	act_remove (id, package, 0);
+	free (path);
+	return 1;
+    } else {
+	DIR *dir;
+	struct dirent *entry;
+	int worked;
+
+	dir = opendir (importdir);
+	if (!dir) {
+	    warning_err ("unable to open %s", importdir);
+	    return 0;
+	}
+	worked = 1;
+	while ((entry = readdir (dir)) != NULL) {
+	    char *importdir_name;
+	    if (!strcmp (entry->d_name, ".") || !strcmp (entry->d_name, ".."))
+		continue;
+	    importdir_name = xasprintf ("%s/%s", importdir, entry->d_name);
+	    if (is_file (importdir_name))
+		worked &= act_unimport (entry->d_name);
+	    free (importdir_name);
+	}
+	closedir (dir);
+	return worked;
+    }
+}
+
 static int act_display (const char *name)
 {
     if (name) {
@@ -766,11 +814,14 @@ static int act_display (const char *name)
 	    package, binfmt->type, binfmt->offset, binfmt->magic, binfmt->mask,
 	    binfmt->interpreter, binfmt->detector);
     } else {
-	struct kvelem *format_iter;
+	gl_list_iterator_t format_iter;
+	const void *format_elem;
 
 	load_all_formats (0);
-	HASH_FOR_EACH (format_iter, formats)
-	    act_display (format_iter->key);
+	format_iter = gl_list_iterator (formats);
+	while (gl_list_iterator_next (&format_iter, &format_elem, NULL))
+	    act_display (((const struct kvelem *) format_elem)->key);
+	gl_list_iterator_free (&format_iter);
     }
     return 1;
 }
@@ -801,6 +852,7 @@ enum opts {
     OPT_INSTALL = 256,
     OPT_REMOVE,
     OPT_IMPORT,
+    OPT_UNIMPORT,
     OPT_DISPLAY,
     OPT_ENABLE,
     OPT_DISABLE,
@@ -828,6 +880,9 @@ static struct argp_option options[] = {
     { "import",		OPT_IMPORT,	0,
 	OPTION_ARG_OPTIONAL | OPTION_HIDDEN,
 	"import packaged format file" },
+    { "unimport",	OPT_UNIMPORT,	0,
+	OPTION_ARG_OPTIONAL | OPTION_HIDDEN,
+	"unimport packaged format file" },
     { "display",	OPT_DISPLAY,	0,
 	OPTION_ARG_OPTIONAL | OPTION_HIDDEN,
 	"display information on binary format" },
@@ -893,6 +948,7 @@ static const char *mode_name (enum opts m)
 	case OPT_INSTALL:	return "install";
 	case OPT_REMOVE:	return "remove";
 	case OPT_IMPORT:	return "import";
+	case OPT_UNIMPORT:	return "unimport";
 	case OPT_DISPLAY:	return "display";
 	case OPT_ENABLE:	return "enable";
 	case OPT_DISABLE:	return "disable";
@@ -917,6 +973,7 @@ static error_t parse_opt (int key, char *arg, struct argp_state *state)
 	case OPT_INSTALL:
 	case OPT_REMOVE:
 	case OPT_IMPORT:
+	case OPT_UNIMPORT:
 	case OPT_DISPLAY:
 	case OPT_ENABLE:
 	case OPT_DISABLE:
@@ -956,6 +1013,7 @@ static error_t parse_opt (int key, char *arg, struct argp_state *state)
 	    return 0;
 
 	case OPT_IMPORT:
+	case OPT_UNIMPORT:
 	case OPT_DISPLAY:
 	case OPT_ENABLE:
 	case OPT_DISABLE:
@@ -1043,8 +1101,8 @@ static error_t parse_opt (int key, char *arg, struct argp_state *state)
 	    if (!mode)
 		argp_error (state,
 			    "you must use one of --install, --remove, "
-			    "--import, --display, --enable, --disable, "
-			    "--find");
+			    "--import, --unimport, --display, --enable, "
+			    "--disable, --find");
 	    else if (mode == OPT_INSTALL) {
 		if (!type)
 		    argp_error (state, "--install requires a <spec> option");
@@ -1064,6 +1122,7 @@ static struct argp argp = {
     "--install <name> <path> <spec>\n"
     "--remove <name> <path>\n"
     "--import [<name>]\n"
+    "--unimport [<name>]\n"
     "--display [<name>]\n"
     "--enable [<name>]\n"
     "--disable [<name>]\n"
@@ -1107,13 +1166,15 @@ int main (int argc, char **argv)
     if (!package)
 	package = ":";
 
-    formats = kvhash_initialize (16, NULL, binfmt_hash_free);
+    formats = kvhash_initialize
+	((gl_listelement_dispose_fn) kvhash_dispose_binfmt);
 
     if (mode == OPT_INSTALL) {
 	struct binfmt *binfmt;
-	Hash_table *format_args;
+	gl_list_t format_args;
 
-	format_args = kvhash_initialize (8, NULL, free);
+	format_args = kvhash_initialize
+	    ((gl_listelement_dispose_fn) kvhash_dispose_plain);
 	kvhash_insert (format_args, "package", package);
 	kvhash_insert (format_args, "type",
 		       (type == OPT_MAGIC) ? "magic" : "extension");
@@ -1133,11 +1194,13 @@ int main (int argc, char **argv)
 #undef ADD_SPEC
 	binfmt = binfmt_new (name, format_args);
 
-	status = act_install (name, binfmt);
+	status = act_install (name, binfmt, 1);
     } else if (mode == OPT_REMOVE)
-	status = act_remove (name, package);
+	status = act_remove (name, package, 1);
     else if (mode == OPT_IMPORT)
 	status = act_import (name);
+    else if (mode == OPT_UNIMPORT)
+	status = act_unimport (name);
     else if (mode == OPT_DISPLAY)
 	status = act_display (name);
     else if (mode == OPT_ENABLE)
